@@ -1,7 +1,7 @@
-/* @(#)getargs.c	2.23 98/03/31 Copyright 1985, 1988, 1995 J. Schilling */
+/* @(#)getargs.c	2.26 99/12/25 Copyright 1985, 1988, 1995 J. Schilling */
 #ifndef lint
 static	char sccsid[] =
-	"@(#)getargs.c	2.23 98/03/31 Copyright 1985, 1988, 1995 J. Schilling";
+	"@(#)getargs.c	2.26 99/12/25 Copyright 1985, 1988, 1995 J. Schilling";
 #endif
 #define	NEW
 /*
@@ -228,14 +228,6 @@ LOCAL int dofile(pac, pav, pargp)
 	if (argp[0] != '-' && argp[0] != '+' && (!checkeql(argp)))
 		return (NOTAFLAG);
 
-	/*
-	 * flags beginning with '-' don't have to include the '-' in
-	 * the format string.
-	 * flags beginning with '+' have to include it in the format string.
-	 */
-	if (argp[0] == '-')
-		(*pargp)++;
-
 	return (NOTAFILE);
 }
 
@@ -261,14 +253,42 @@ LOCAL int doflag(pac, pav, argp, fmt, setargs, oargs)
 	long	val;
 	int	singlecharflag	= 0;
 	BOOL	isspec;
+	BOOL	hasdash		= FALSE;
+	BOOL	doubledash	= FALSE;
 	BOOL	haseql		= checkeql(argp);
-	const char	*sargp	= argp;
+	const char	*sargp;
 	const char	*sfmt	= fmt;
 	va_list	args;
 	char	*const	*spav	= *pav;
 	int		spac	= *pac;
 	void		*curarg	= (void *)0;
 
+	/*
+	 * flags beginning with '-' don't have to include the '-' in
+	 * the format string.
+	 * flags beginning with '+' have to include it in the format string.
+	 */
+	if (argp[0] == '-') {
+		argp++;
+		hasdash = TRUE;
+		/*
+		 * Implement legacy support for --longopt
+		 * If we find a double dash, we do not look for combinations
+		 * of boolean single char flags.
+		 */
+		if (argp[0] == '-') {
+			argp++;
+			doubledash = TRUE;
+			/*
+			 * Allow -- only for long options.
+			 */
+			if (argp[1] == '\0') {
+				return (BADFLAG);
+			}
+		}
+	}
+	sargp = argp;
+
 	/*
 	 * Initialize 'args' to the start of the argument list.
 	 * I don't know any portable way to copy an arbitrary
@@ -345,12 +365,16 @@ LOCAL int doflag(pac, pav, argp, fmt, setargs, oargs)
 					 * and we only allow to match
 					 * the special pattern '&'.
 					 * We need this e.g. for 'make'.
-					 * Here allow any flag type argument to
+					 * We allow any flag type argument to
 					 * match the format string "&" to set
 					 * up a function that handles all odd
 					 * stuff that getargs will not grok.
+					 * In addition, to allow getargs to be
+					 * used for CPP type flags we allow to
+					 * match -Dabc=xyz on 'D&'. Note that
+					 * Dabc=xyz will not match 'D&'.
 					 */
-					if (argp != sargp || *fmt != '&')
+					if ((!hasdash && argp != sargp) || *fmt != '&')
 						goto nextarg;
 				}
 				/*
@@ -401,7 +425,7 @@ LOCAL int doflag(pac, pav, argp, fmt, setargs, oargs)
 			/*
 			 * Boolean type has been tested before.
 			 */
-			if (singlecharflag && 
+			if (singlecharflag && !doubledash &&
 			   (val = dosflags(sargp, sfmt, setargs, oargs)) !=
 								BADFLAG)
 				return (val);
@@ -677,7 +701,9 @@ LOCAL int dosflags(argp, fmt, setargs, oargs)
 LOCAL int checkfmt(fmt)
 	const char	*fmt;
 {
-	char c = *(++fmt);
+	char	c;
+
+	c = *(++fmt);	/* non constant expression */
 
 
 	if (c == ',' || c == '\0') {
